import logging

import mopidy_somafm
import pykka
import requests
import configparser
import random
from mopidy import backend, httpclient
from mopidy.models import Album, Artist, Image, Ref, Track

from .somafm import SomaFMClient, extract_somafm_channel_name_from_uri

logger = logging.getLogger(__name__)


class SomaFMBackend(pykka.ThreadingActor, backend.Backend):
    def __init__(self, config, audio):
        super().__init__()

        user_agent = "{}/{}".format(
            mopidy_somafm.Extension.dist_name, mopidy_somafm.__version__
        )

        self.somafm = SomaFMClient(config["proxy"], user_agent)
        self.library = SomaFMLibraryProvider(backend=self)
        self.playback = SomaFMPlayback(
            audio=audio,
            backend=self,
            proxy_config=config["proxy"],
            user_agent=user_agent,
        )

        self.uri_schemes = ["somafm"]
        self.quality = config["somafm"]["quality"]
        self.encoding = config["somafm"]["encoding"]
        self.dj_as_artist = config["somafm"]["dj_as_artist"]

    def on_start(self):
        self.somafm.refresh(self.encoding, self.quality)


class SomaFMLibraryProvider(backend.LibraryProvider):

    root_directory = Ref.directory(uri="somafm:root", name="SomaFM")

    def lookup(self, uri):
        # Whatever the uri, it will always contains one track
        # which is a url to a pls
        channel_name = extract_somafm_channel_name_from_uri(uri)
        if channel_name is None:
            return None

        channel_data = self.backend.somafm.channels[channel_name]

        # Artists
        if self.backend.dj_as_artist:
            artist = Artist(name=channel_data["dj"])
        else:
            artist = Artist()

        # Build album (idem as playlist, but with more metada)
        album = Album(
            artists=[artist],
            name=channel_data["title"],
        )

        track = Track(
            artists=[artist],
            album=album,
            last_modified=channel_data["updated"],
            comment=channel_data["description"],
            genre=channel_data["genre"],
            name=channel_data["title"],
            uri="somafm:channel:/%s" % (channel_name),
        )

        return [track]

    def browse(self, uri):

        if uri != "somafm:root":
            return []

        result = []
        for channel in self.backend.somafm.channels:
            result.append(
                Ref.track(
                    uri="somafm:channel:/%s" % (channel),
                    name=self.backend.somafm.channels[channel]["title"],
                )
            )

        result.sort(key=lambda ref: ref.name.lower())
        return result

    def get_images(self, uris):

        images = {}

        for uri in uris:
            channel_name = extract_somafm_channel_name_from_uri(uri)
            if channel_name is not None:
                image = Image(uri=self.backend.somafm.images[channel_name])
                images[uri] = [image]

        return images


class SomaFMPlayback(backend.PlaybackProvider):
    def __init__(self, audio, backend, proxy_config=None, user_agent=None):
        super().__init__(audio=audio, backend=backend)

        # Build requests session
        self.session = requests.Session()
        if proxy_config is not None:
            proxy = httpclient.format_proxy(proxy_config)
            self.session.proxies.update({"http": proxy, "https": proxy})

        full_user_agent = httpclient.format_user_agent(user_agent)
        self.session.headers.update({"user-agent": full_user_agent})

    def translate_uri(self, uri):
        try:
            channel_name = extract_somafm_channel_name_from_uri(uri)
            if channel_name is None:
                return None

            channel_data = self.backend.somafm.channels.get(channel_name)

            r = self.session.get(channel_data["pls"])
            if r.status_code != 200:
                return None

            pls = configparser.ConfigParser()
            pls.read_string(r.text)
            playlist = pls["playlist"]
            num = int(playlist["numberofentries"])
            return playlist["File" + str(random.randint(1, num))]

        except Exception:
            return None



import logging
import pathlib

import pkg_resources

from mopidy import config, ext

__version__ = pkg_resources.get_distribution("Mopidy-SomaFM").version

logger = logging.getLogger(__name__)


class Extension(ext.Extension):

    dist_name = "Mopidy-SomaFM"
    ext_name = "somafm"
    version = __version__

    def get_default_config(self):
        return config.read(pathlib.Path(__file__).parent / "ext.conf")

    def get_config_schema(self):
        schema = super().get_config_schema()
        schema["encoding"] = config.String(choices=("aac", "mp3", "aacp"))
        schema["quality"] = config.String(
            choices=("highest", "fast", "slow", "firewall")
        )
        schema["dj_as_artist"] = config.Boolean()
        return schema

    def setup(self, registry):
        from .backend import SomaFMBackend

        registry.add("backend", SomaFMBackend)



import collections
import logging
import re
from urllib.parse import urlsplit

import requests

from mopidy import httpclient

try:
    import xml.etree.cElementTree as ET
except ImportError:
    import xml.etree.ElementTree as ET


logger = logging.getLogger(__name__)

#
#  Channels are playlist and Album
#  PLS are tracks
#  PLS contents for internal use
#


def extract_somafm_channel_name_from_uri(uri):

    if not uri.startswith("somafm:"):
        return None

    channel_separator = uri.find("/")
    if channel_separator == -1:
        return None

    return uri[channel_separator + 1 :]


class SomaFMClient:

    CHANNELS_URI = "https://api.somafm.com/channels.xml"

    # All channels seem to have this combination of quality/encoding available
    FALLBACK_QUALITY = "fast"
    FALLBACK_ENCODING = "mp3"

    channels = {}
    images = {}

    def __init__(self, proxy_config=None, user_agent=None):
        super().__init__()

        # Build requests session
        self.session = requests.Session()
        if proxy_config is not None:
            proxy = httpclient.format_proxy(proxy_config)
            self.session.proxies.update({"http": proxy, "https": proxy})

        full_user_agent = httpclient.format_user_agent(user_agent)
        self.session.headers.update({"user-agent": full_user_agent})

    def refresh(self, encoding, quality):
        # clean previous data
        self.channels = {}

        # download channels xml file
        channels_content = self._downloadContent(self.CHANNELS_URI)
        if channels_content is None:
            logger.error("Cannot fetch %s" % (self.CHANNELS_URI))
            return

        # parse XML
        root = ET.fromstring(channels_content)

        for child_channel in root:

            pls_id = child_channel.attrib["id"]
            channel_data = {}
            channel_all_pls = collections.defaultdict(dict)

            for child_detail in child_channel:

                key = child_detail.tag
                val = child_detail.text

                if key in ["title", "image", "dj", "genre", "description"]:
                    channel_data[key] = val
                elif key == "updated":
                    channel_data["updated"] = int(val)
                elif "pls" in key:
                    pls_quality = key[:-3]
                    pls_format = child_detail.attrib["format"]

                    channel_all_pls[pls_quality][pls_format] = val

                    # firewall playlist are fastpls+mp3 but with fw path
                    if pls_quality == "fast" and pls_format == "mp3":
                        r1 = urlsplit(val)
                        channel_all_pls["firewall"][
                            "mp3"
                        ] = "{}://{}/{}".format(
                            r1.scheme, r1.netloc, "fw" + r1.path
                        )

            channel_pls = self._choose_pls(channel_all_pls, encoding, quality)

            if channel_pls is not None:
                channel_data["pls"] = channel_pls
                self.channels[pls_id] = channel_data
                self.images[pls_id] = channel_data["image"]

        logger.info("Loaded %i SomaFM channels" % (len(self.channels)))

    def extractStreamUrlFromPls(self, pls_uri):
        pls_content = self._downloadContent(pls_uri)
        if pls_content is None:
            logger.error("Cannot fetch %s" % (pls_uri))
            return pls_uri

        # try to find FileX=<stream url>
        try:
            m = re.search(r"^(File\d)=(?P<stream_url>\S+)", pls_content, re.M)
            if m:
                return m.group("stream_url")
            else:
                return pls_uri
        except BaseException:
            return pls_uri

    def _choose_pls(self, all_pls, encoding, quality):
        if not all_pls:
            return None

        if quality in all_pls:
            quality_pls = all_pls[quality]
        elif self.FALLBACK_QUALITY in all_pls:
            quality_pls = all_pls[self.FALLBACK_QUALITY]
        else:
            quality_pls = all_pls[next(iter(all_pls))]

        if not quality_pls:
            return None

        if encoding in quality_pls:
            pls = quality_pls[encoding]
        elif self.FALLBACK_ENCODING in all_pls:
            pls = quality_pls[self.FALLBACK_ENCODING]
        else:
            pls = quality_pls[next(iter(quality_pls))]

        return pls

    def _downloadContent(self, url):
        try:
            r = self.session.get(url)
            logger.debug("Get %s : %i", url, r.status_code)

            if r.status_code != 200:
                logger.error(
                    "SomaFM: %s is not reachable [http code:%i]",
                    url,
                    r.status_code,
                )
                return None

        except requests.exceptions.RequestException as e:
            logger.error("SomaFM RequestException: %s", e)
        except requests.exceptions.ConnectionError as e:
            logger.error("SomaFM ConnectionError: %s", e)
        except requests.exceptions.URLRequired as e:
            logger.error("SomaFM URLRequired: %s", e)
        except requests.exceptions.TooManyRedirects as e:
            logger.error("SomaFM TooManyRedirects: %s", e)
        except Exception as e:
            logger.error("SomaFM exception: %s", e)
        else:
            return r.text

        return None



import unittest

from mopidy_somafm.somafm import SomaFMClient


class SomaFMClientTest(unittest.TestCase):
    def test_refresh(self):
        sfmc = SomaFMClient()
        sfmc.refresh("mp3", "fast")

        self.assertIsNotNone(sfmc.channels)
        self.assertNotEqual(len(sfmc.channels), 0)

    def test_refresh_firewall(self):
        sfmc = SomaFMClient()
        sfmc.refresh("mp3", "firewall")

        self.assertIsNotNone(sfmc.channels)
        self.assertNotEqual(len(sfmc.channels), 0)

    def test_refresh_no_channels(self):
        sfmc = SomaFMClient()
        sfmc.CHANNELS_URI = ""
        sfmc.refresh("mp3", "fast")

        self.assertDictEqual(sfmc.channels, {})
        self.assertEqual(len(sfmc.channels), 0)

    def test_downloadContent(self):
        url = "http://api.somafm.com/channels.xml"
        sfmc = SomaFMClient()
        data = sfmc._downloadContent(url)
        self.assertNotEqual(len(data), 0)

    def test_extractStreamUrlFromPls(self):
        url = "http://somafm.com/groovesalad.pls"
        sfmc = SomaFMClient()
        data = sfmc.extractStreamUrlFromPls(url)
        self.assertNotEqual(len(data), 0)
        self.assertNotEqual(data, url)

    def test_extractStreamUrlFromPls_unknown(self):
        url = "http://somafm.com/noneazerty.pls"
        sfmc = SomaFMClient()
        data = sfmc.extractStreamUrlFromPls(url)
        self.assertNotEqual(len(data), 0)
        self.assertEqual(data, url)



from mopidy_somafm import Extension


def test_get_default_config():
    ext = Extension()

    config = ext.get_default_config()

    assert "[somafm]" in config
    assert "enabled = true" in config


def test_get_config_schema():
    ext = Extension()

    schema = ext.get_config_schema()

    assert "quality" in schema
    assert "encoding" in schema
